import unittest
from osm_export_tool.mapping import Mapping
from osm_export_tool import GeomType

class TestMapping(unittest.TestCase):
    def test_basic_mapping(self):
        y = '''
        buildings:
          select:
            - name
        '''
        m = Mapping(y)
        self.assertEqual(len(m.themes),1)
        theme = m.themes[0]
        self.assertEqual(theme.name,'buildings')
        self.assertTrue(theme.points)
        self.assertTrue(theme.lines)
        self.assertTrue(theme.polygons)
        self.assertTrue('name' in theme.keys)

    def test_geom_types(self):
        y = '''
        buildings:
          types:
            - points
          select:
            - name
        '''
        m = Mapping(y)
        self.assertTrue(m.themes[0].points)
        self.assertFalse(m.themes[0].lines)
        self.assertFalse(m.themes[0].polygons)
        y = '''
        buildings:
          types:
            - lines
          select:
            - name
        '''
        m = Mapping(y)
        self.assertFalse(m.themes[0].points)
        self.assertTrue(m.themes[0].lines)
        self.assertFalse(m.themes[0].polygons)
        y = '''
        buildings:
          types:
            - polygons
          select:
            - name
        '''
        m = Mapping(y)
        self.assertFalse(m.themes[0].points)
        self.assertFalse(m.themes[0].lines)
        self.assertTrue(m.themes[0].polygons)
        

    def test_key_selections(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - addr:housenumber
        '''
        m = Mapping(y)
        self.assertTrue('addr:housenumber' in m.themes[0].keys)

    def test_where(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - addr:housenumber
          where:
            - building = 'yes'
        '''
        m = Mapping(y)
        self.assertFalse(m.themes[0].matches(GeomType.POINT,{'building':'yes'}))
        self.assertFalse(m.themes[0].matches(GeomType.POLYGON,{'building':'no'}))
        self.assertTrue(m.themes[0].matches(GeomType.POLYGON,{'building':'yes'}))

    def test_default_matcher(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - addr:housenumber
        '''
        m = Mapping(y)
        self.assertTrue(m.themes[0].matches(GeomType.POLYGON,{'addr:housenumber':'1234'}))

    def test_multiple_matchers(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - addr:housenumber
          where: 
            - building = 'yes'
            - amenity = 'parking'
        '''
        m = Mapping(y)
        self.assertTrue(m.themes[0].matches(GeomType.POLYGON,{'building':'yes'}))
        self.assertTrue(m.themes[0].matches(GeomType.POLYGON,{'amenity':'parking'}))

    def test_nonlist_matcher(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - addr:housenumber
          where: building = 'yes'
        '''
        m = Mapping(y)
        self.assertTrue(m.themes[0].matches(GeomType.POLYGON,{'building':'yes'}))

    def test_gt(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - building
          where: height > 20
        '''
        m = Mapping(y)
        self.assertTrue(m.themes[0].matches(GeomType.POLYGON,{'height':21}))
        self.assertFalse(m.themes[0].matches(GeomType.POLYGON,{'height':20}))

    def test_default_osm_id(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - building
        '''
        m = Mapping(y)
        self.assertTrue(m.themes[0].osm_id)
        self.assertFalse('osm_id' in m.themes[0].keys)
        m = Mapping(y,default_osm_id=False)
        self.assertFalse(m.themes[0].osm_id)

    def test_osm_id_override(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - building
            - osm_id
        '''
        m = Mapping(y)
        self.assertTrue(m.themes[0].osm_id)
        self.assertFalse('osm_id' in m.themes[0].keys)
        m = Mapping(y,default_osm_id=False)
        self.assertTrue(m.themes[0].osm_id)
        self.assertFalse('osm_id' in m.themes[0].keys)

    def test_duplicate_key(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - building
            - building
        '''
        m = Mapping(y)
        self.assertTrue(len(m.themes[0].keys) == 1)

    def test_extra(self):
        y = '''
        buildings:
          foo:
            bar: baz
          select:
            - building
        '''
        m = Mapping(y)
        self.assertEqual(m.themes[0].extra,{'foo':{'bar':'baz'}})

class TestMappingValidation(unittest.TestCase):
    def test_empty_yaml(self):
        y = '''
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_bad_yaml(self):
        y = '''
        buildings
          types:
            - polygons
          select:
            - building
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_no_select(self):
        y = '''
        buildings:
          types:
            - polygons
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_invalid_type(self):
        y = '''
        buildings:
          types:
            - polygon
          select:
            - building
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_empty_sql(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - building
          where:
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_invalid_sql(self):
        y = '''
        buildings:
          types:
            - polygons
          select:
            - building
          where: XXX aaa
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_wrong_yaml_list(self):
        y = '''
        buildings:
          types: polygons
          select:
            - building
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_invalid_yaml_parse(self):
        y = '''
        buildings:
           - types: 
             - polygons
          select:
            - building
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)

    def test_wrong_yaml_parse(self):
        y = '''
        buildings:
          - types: 
            - polygons
          - select:
            - building
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)
        self.assertTrue('must be YAML dict' in errors[0])

    def test_wrong_toplevel_themes(self):
        y = '''
         - buildings:
            types: 
            - polygons
            select:
            - building
        '''
        m, errors = Mapping.validate(y)
        self.assertTrue(m is None)
        self.assertTrue(len(errors) == 1)



import os
import sys
import time
import argparse
import osm_export_tool.tabular as tabular
import osm_export_tool.nontabular as nontabular
from osm_export_tool.mapping import Mapping
from osm_export_tool.geometry import load_geometry

def main():
	parser = argparse.ArgumentParser(description='Export OSM data in other file formats.')
	parser.add_argument('osm_file',  help='OSM .pbf or .xml input file')
	parser.add_argument('output_name', help='Output prefix')
	parser.add_argument('-f','--formats', dest='formats',default='gpkg',help='List of formats e.g. gpkg,shp,kml (default: gpkg)')
	parser.add_argument('-m','--mapping', dest='mapping',help='YAML mapping of features schema. If not specified, a default is used.')
	parser.add_argument('--clip', dest='clip',help='GeoJSON or POLY file to clip geometries.')
	parser.add_argument('-v','--verbose', action='store_true')
	parser.add_argument('--omit-osm-ids', action='store_true')
	parsed = parser.parse_args()

	mapping_txt = None
	if parsed.mapping:
		with open(parsed.mapping,'r') as f:
			mapping_txt = f.read()
	else:
		default_mapping = os.path.join(os.path.dirname(__file__), 'mappings/default.yml')
		with open(default_mapping,'r') as f:
			mapping_txt = f.read()
	
	mapping = Mapping(mapping_txt,default_osm_id=not parsed.omit_osm_ids)

	clipping_geom = None
	if parsed.clip:
		with open(parsed.clip,'r') as f:
			clipping_geom = load_geometry(f.read())

	formats = parsed.formats.split(',')
	
	tabular_outputs = []
	if 'gpkg' in formats:
		tabular_outputs.append(tabular.Geopackage(parsed.output_name,mapping))
	if 'shp' in formats:
		tabular_outputs.append(tabular.Shapefile(parsed.output_name,mapping))
	if 'kml' in formats:
		tabular_outputs.append(tabular.Kml(parsed.output_name,mapping))

	nontabular_outputs = []
	if 'mwm' in formats:
		nontabular_outputs.append(nontabular.Omim())
	if 'img' in formats:
		nontabular_outputs.append(nontabular.GarminIMG())
	if 'obf' in formats:
		nontabular_outputs.append(nontabular.Osmand())

	if len(tabular_outputs) > 0:
		h = tabular.Handler(tabular_outputs,mapping,clipping_geom=clipping_geom)
		start_time = time.time()
		h.apply_file(parsed.osm_file, locations=True, idx='sparse_file_array')

		for output in tabular_outputs:
			output.finalize()
		print('Completed in {0} seconds.'.format(time.time() - start_time))

		for output in tabular_outputs:
			for file in output.files:
				print(file)

	if len(nontabular_outputs) > 0:
		for output in nontabular_outputs:
			output.run(pbf,tmpdir)


import json
from shapely.geometry import shape, MultiPolygon, Polygon

def parse_poly(lines):
    """ Parse an Osmosis polygon filter file.
        Accept a sequence of lines from a polygon file, return a shapely.geometry.MultiPolygon object.
        https://wiki.openstreetmap.org/wiki/Osmosis/Polygon_Filter_File_Python_Parsing
    """
    in_ring = False
    coords = []
    for (index, line) in enumerate(lines):
        if index == 0:
            # first line is junk.
            continue
        elif index == 1:
            # second line is the first polygon ring.
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True
        elif in_ring and line.strip() == 'END':
            # we are at the end of a ring, perhaps with more to come.
            in_ring = False
        elif in_ring:
            # we are in a ring and picking up new coordinates.
            ring.append(list(map(float, line.split())))
        elif not in_ring and line.strip() == 'END':
            # we are at the end of the whole polygon.
            break
        elif not in_ring and line.startswith('!'):
            # we are at the start of a polygon part hole.
            coords[-1][1].append([])
            ring = coords[-1][1][-1]
            in_ring = True
        elif not in_ring:
            # we are at the start of a polygon part.
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True
    
    return MultiPolygon(coords)

def load_geometry(txt):
	try:
		j = json.loads(txt)
		if j['type'] == 'FeatureCollection':
			print("Warning: using first feature of --clip FeatureCollection.")
			return shape(j['features'][0]['geometry'])
		else:
			return shape(j)
	except json.decoder.JSONDecodeError:
		pass
	return parse_poly(txt.split('\n'))



import os
from enum import Enum

name = 'osm_export_tool'

class GeomType(Enum):
    POINT = 1
    LINE = 2
    POLYGON = 3

def GetHumanReadable(size,precision=2):
    suffixes=['B','KB','MB','GB','TB']
    suffixIndex = 0
    while size > 1024 and suffixIndex < 4:
        suffixIndex += 1 #increment the index of the suffix
        size = size/1024.0 #apply the division
    return "%.*f%s"%(precision,size,suffixes[suffixIndex])

# can be more than one file (example: Shapefile w/ sidecars)
class File:
    def __init__(self,output_name,parts,extra = {}):
        self.output_name = output_name
        self.parts = parts
        self.extra = extra

    @classmethod
    def shp(cls,name,extra = {}):
        parts = [name + '.shp']
        parts.append(name + '.shx')
        parts.append(name + '.prj')
        parts.append(name + '.cpg')
        parts.append(name + '.dbf')
        return cls('shp',parts,extra)

    def size(self):
        total = 0
        for part in self.parts:
            total = total + os.path.getsize(part)
        return total

    def __str__(self):
        return '{0} {1} {2} {3}'.format(self.output_name,self.extra,','.join(self.parts),GetHumanReadable(self.size()))

    def __repr__(self):
        return self.__str__()


from base64 import b64decode
import os
import re

import osmium as o
import ogr
from shapely.wkb import loads, dumps

from osm_export_tool import GeomType, File

fab = o.geom.WKBFactory()
create_geom = lambda b : ogr.CreateGeometryFromWkb(bytes.fromhex(b))
epsg_4326 = ogr.osr.SpatialReference()
epsg_4326.ImportFromEPSG(4326)

CLOSED_WAY_KEYS = ['aeroway','amenity','boundary','building','building:part','craft','geological','historic','landuse','leisure','military','natural','office','place','shop','sport','tourism']
CLOSED_WAY_KEYVALS = {'highway':'platform','public_transport':'platform'}
def closed_way_is_polygon(tags):
    for key in CLOSED_WAY_KEYS:
        if key in tags:
            return True
    for key, val in CLOSED_WAY_KEYVALS.items():
        if key in tags and tags[key] == val:
            return True
    return False

def make_filename(s):
    return s.lower().replace(' ','_')

class Kml:
    class Layer:
        def __init__(self,driver,file_name,ogr_geom_type,theme):
            self.columns = theme.keys
            self.ds = driver.CreateDataSource(file_name + '.kml')
            self.ogr_layer = self.ds.CreateLayer(theme.name, epsg_4326, ogr_geom_type)

            if theme.osm_id:
                self.osm_id = True
                field_name = ogr.FieldDefn('osm_id', ogr.OFTInteger64)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)
            else:
                self.osm_id = False

            for column in self.columns:
                field_name = ogr.FieldDefn(column, ogr.OFTString)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)

            self.defn = self.ogr_layer.GetLayerDefn()

    def __init__(self,output_name,mapping):
        driver = ogr.GetDriverByName('KML')

        self.files = []
        self.layers = {}
        for t in mapping.themes:
            name = output_name + '_' + make_filename(t.name)
            if t.points:
                self.layers[(t.name,GeomType.POINT)] = Kml.Layer(driver,name + '_points',ogr.wkbPoint,t)
                self.files.append(File('kml',[name + '_points.kml'],{'theme':t.name}))
            if t.lines:
                self.layers[(t.name,GeomType.LINE)] = Kml.Layer(driver,name + '_lines',ogr.wkbLineString,t)
                self.files.append(File('kml',[name + '_lines.kml'],{'theme':t.name}))
            if t.polygons:
                self.layers[(t.name,GeomType.POLYGON)] = Kml.Layer(driver,name + '_polygons',ogr.wkbMultiPolygon,t)
                self.files.append(File('kml',[name + '_polygons.kml'],{'theme':t.name}))

    def write(self,osm_id,layer_name,geom_type,geom,tags):
        layer = self.layers[(layer_name,geom_type)]
        feature = ogr.Feature(layer.defn)
        feature.SetGeometry(geom)
        if layer.osm_id:
            feature.SetField('osm_id',osm_id)
        for col in layer.columns:
            if col in tags:
                feature.SetField(col,tags[col])
        layer.ogr_layer.CreateFeature(feature)

    def finalize(self):
        self.layers = None
        self.ds = None

class Shapefile:
    class Layer:
        def __init__(self,driver,file_name,ogr_geom_type,theme):
            def launderName(col):
                return re.sub(r'[^a-zA-Z0-9_]', '', col)[0:10]

            self.columns = theme.keys
            self.ds = driver.CreateDataSource(file_name + '.shp')
            self.ogr_layer = self.ds.CreateLayer(theme.name, epsg_4326, ogr_geom_type,options=['ENCODING=UTF-8'])

            if theme.osm_id:
                self.osm_id = True
                field_name = ogr.FieldDefn('osm_id', ogr.OFTInteger64)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)
            else:
                self.osm_id = False

            self.launderedNames = {}
            for column in self.columns:
                laundered_name = launderName(column)
                field_name = ogr.FieldDefn(laundered_name, ogr.OFTString)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)
                self.launderedNames[column] = laundered_name

            self.defn = self.ogr_layer.GetLayerDefn()

    def __init__(self,output_name,mapping):
        driver = ogr.GetDriverByName('ESRI Shapefile')

        self.files = []
        self.layers = {}
        for t in mapping.themes:
            name = output_name + '_' + make_filename(t.name)
            if t.points:
                self.layers[(t.name,GeomType.POINT)] = Shapefile.Layer(driver,name + '_points',ogr.wkbPoint,t)
                self.files.append(File.shp(name + '_points',{'theme':t.name}))
            if t.lines:
                self.layers[(t.name,GeomType.LINE)] = Shapefile.Layer(driver,name + '_lines',ogr.wkbLineString,t)
                self.files.append(File.shp(name + '_lines',{'theme':t.name}))
            if t.polygons:
                self.layers[(t.name,GeomType.POLYGON)] = Shapefile.Layer(driver,name + '_polygons',ogr.wkbMultiPolygon,t)
                self.files.append(File.shp(name + '_polygons',{'theme':t.name}))

    def write(self,osm_id,layer_name,geom_type,geom,tags):
        layer = self.layers[(layer_name,geom_type)]
        feature = ogr.Feature(layer.defn)
        feature.SetGeometry(geom)
        if layer.osm_id:
            feature.SetField('osm_id',osm_id)
        for col in layer.columns:
            if col in tags:
                feature.SetField(layer.launderedNames[col],tags[col])
        layer.ogr_layer.CreateFeature(feature)

    def finalize(self):
        self.layers = None
        self.ds = None

class Geopackage:
    class Layer:
        def __init__(self,ds,theme):
            self.ogr_layer = ds.CreateLayer(theme.name, epsg_4326, ogr.wkbUnknown,options=['SPATIAL_INDEX=NO'])

            if theme.osm_id:
                self.osm_id = True
                field_name = ogr.FieldDefn('osm_id', ogr.OFTInteger64)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)
            else:
                self.osm_id = False

            self.columns = theme.keys
            for column_name in self.columns:
                field_name = ogr.FieldDefn(column_name, ogr.OFTString)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)
            self.defn = self.ogr_layer.GetLayerDefn()

    def __init__(self,output_name,mapping):
        driver = ogr.GetDriverByName('GPKG')
        self.ds = driver.CreateDataSource(output_name + '.gpkg')
        self.ds.StartTransaction()

        self.files = [File('gpkg',[output_name + '.gpkg'])]
        self.layers = {}
        for theme in mapping.themes:
            layer = Geopackage.Layer(self.ds,theme)
            if theme.points:
                self.layers[(theme.name,GeomType.POINT)] = layer
            if theme.lines:
                self.layers[(theme.name,GeomType.LINE)] = layer
            if theme.polygons:
                self.layers[(theme.name,GeomType.POLYGON)] = layer

    def write(self,osm_id,layer_name,geom_type,geom,tags):
        layer = self.layers[(layer_name,geom_type)]
        feature = ogr.Feature(layer.defn)
        feature.SetGeometry(geom)
        if layer.osm_id:
            feature.SetField('osm_id',osm_id)
        for column_name in layer.columns:
            if column_name in tags:
                feature.SetField(column_name,tags[column_name])
        layer.ogr_layer.CreateFeature(feature)

    def finalize(self):
        self.ds.CommitTransaction()
        self.layers = None
        self.ds = None

# special case where each theme is a separate geopackage, for legacy reasons
class MultiGeopackage:
    class Layer:
        def __init__(self,output_name,theme):
            driver = ogr.GetDriverByName('GPKG')
            self.ds = driver.CreateDataSource(output_name + '_' + make_filename(theme.name) + '.gpkg')
            self.ds.StartTransaction()
            self.ogr_layer = self.ds.CreateLayer(theme.name, epsg_4326, ogr.wkbUnknown,options=['SPATIAL_INDEX=NO'])

            if theme.osm_id:
                self.osm_id = True
                field_name = ogr.FieldDefn('osm_id', ogr.OFTInteger64)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)
            else:
                self.osm_id = False

            self.columns = theme.keys
            for column_name in self.columns:
                field_name = ogr.FieldDefn(column_name, ogr.OFTString)
                field_name.SetWidth(254)
                self.ogr_layer.CreateField(field_name)
            self.defn = self.ogr_layer.GetLayerDefn()

    def __init__(self,output_name,mapping):
        self.files = []
        self.layers = {}
        for theme in mapping.themes:
            layer = MultiGeopackage.Layer(output_name, theme)
            self.files.append(File('gpkg',[output_name + '_' + make_filename(theme.name) + '.gpkg'],{'theme':theme.name}))
            if theme.points:
                self.layers[(theme.name,GeomType.POINT)] = layer
            if theme.lines:
                self.layers[(theme.name,GeomType.LINE)] = layer
            if theme.polygons:
                self.layers[(theme.name,GeomType.POLYGON)] = layer

    def write(self,osm_id,layer_name,geom_type,geom,tags):
        layer = self.layers[(layer_name,geom_type)]
        feature = ogr.Feature(layer.defn)
        feature.SetGeometry(geom)
        if layer.osm_id:
            feature.SetField('osm_id',osm_id)
        for column_name in layer.columns:
            if column_name in tags:
                feature.SetField(column_name,tags[column_name])
        layer.ogr_layer.CreateFeature(feature)

    def finalize(self):
        for k, layer in self.layers.items():
            layer.ds.CommitTransaction()
        self.layers = None

class Handler(o.SimpleHandler):
    def __init__(self,outputs,mapping,clipping_geom=None):
        super(Handler, self).__init__()
        self.outputs = outputs
        self.mapping = mapping
        self.clipping_geom = clipping_geom

    def node(self,n):
        if len(n.tags) == 0:
            return
        geom = None
        for theme in self.mapping.themes:
            if theme.matches(GeomType.POINT,n.tags):
                if not geom:
                    wkb = fab.create_point(n)
                    if self.clipping_geom:
                        sg = loads(bytes.fromhex(wkb))
                        if not self.clipping_geom.intersects(sg):
                            return
                    geom = create_geom(wkb)
                for output in self.outputs:
                    output.write(n.id,theme.name,GeomType.POINT,geom,n.tags)

    def way(self, w):
        if len(w.tags) == 0:
            return
        if w.is_closed() and closed_way_is_polygon(w.tags): # this will be handled in area()
            return
        try:
            # NOTE: it is possible this is actually a MultiLineString
            # in the case where a LineString is clipped by the clipping geom,
            # or the way is self-intersecting
            # but GDAL and QGIS seem to handle it OK.
            linestring = None
            for theme in self.mapping.themes:
                if theme.matches(GeomType.LINE,w.tags):
                    if not linestring:
                        wkb = fab.create_linestring(w)
                        if self.clipping_geom:
                            sg = loads(bytes.fromhex(wkb))
                            sg = self.clipping_geom.intersection(sg)
                            if sg.is_empty:
                                return
                            linestring = ogr.CreateGeometryFromWkb(dumps(sg))
                        else:
                            linestring = create_geom(wkb)
                    for output in self.outputs:
                        output.write(w.id,theme.name,GeomType.LINE,linestring,w.tags)
        except RuntimeError:
            print("Incomplete way: {0}".format(w.id))

    def area(self,a):
        if len(a.tags) == 0:
            return
        if not closed_way_is_polygon(a.tags):
            return
        osm_id = a.orig_id() if a.from_way() else -a.orig_id()
        try:
            multipolygon = None
            for theme in self.mapping.themes:
                if theme.matches(GeomType.POLYGON,a.tags):
                    if not multipolygon:
                        wkb = fab.create_multipolygon(a)
                        if self.clipping_geom:
                            sg = loads(bytes.fromhex(wkb))
                            sg = self.clipping_geom.intersection(sg)
                            if sg.is_empty:
                                return
                            multipolygon = ogr.CreateGeometryFromWkb(dumps(sg))
                        else:
                            multipolygon = create_geom(wkb)
                    for output in self.outputs:
                        output.write(osm_id,theme.name,GeomType.POLYGON,multipolygon,a.tags)
        except RuntimeError:
            print('Invalid area: {0}'.format(a.orig_id()))





import json
import os
import shutil
import subprocess
import requests
from string import Template
from osm_export_tool.sql import to_prefix
import shapely.geometry

# path must return a path to an .osm.pbf or .osm.xml on the filesystem

class Pbf:
    def __init__(self,path):
        self._path = path

    def fetch(self):
        pass

    def path(self):
        return self._path

class OsmExpress:
    def __init__(self,osmx_path,db_path,geom,output_path,use_existing=True,tempdir=None):
        self.osmx_path = osmx_path
        self.db_path = db_path
        self.geom = geom
        self.output_path = output_path
        self.use_existing = use_existing
        self.tempdir = tempdir

    def fetch(self):
        region_json = os.path.join(self.tempdir,'region.json')
        with open(region_json,'w') as f:
            f.write(json.dumps(shapely.geometry.mapping(self.geom)))
        subprocess.check_call([self.osmx_path,'extract',self.db_path,self.output_path,'--region',region_json])
        os.remove(region_json)

    def path(self):
        if os.path.isfile(self.output_path) and self.use_existing:
            return self.output_path
        else:
            self.fetch()
        return self.output_path

class Overpass:
    @classmethod
    def filters(cls,mapping):
        nodes = set()
        ways = set()
        relations = set()
        for t in mapping.themes:
            parts = cls.parts(t.matcher.expr)
            if t.points:
                for part in parts:
                    nodes.add(part)
            if t.lines:
                for part in parts:
                    ways.add(part)
            if t.polygons:
                for part in parts:
                    ways.add(part)
                    relations.add(part)
        return nodes,ways,relations

    # force quoting of strings to handle keys with colons
    @classmethod
    def parts(cls, expr):
        def _parts(prefix):
            op = prefix[0]
            if op == '=':
                return ["['{0}'='{1}']".format(prefix[1],prefix[2])]
            if op in ['<','>','<=','>='] or op == 'notnull':
                return ["['{0}']".format(prefix[1])]
            if op == 'in':
                x = "['{0}'~'{1}']".format(prefix[1],'|'.join(prefix[2]))
                return [x]
            if op == 'and' or op == 'or':
                return _parts(prefix[1]) + _parts(prefix[2])
        return _parts(expr)

    @classmethod
    def sql(cls,str):
        return cls.parts(to_prefix(str))

    def __init__(self,hostname,geom,path,use_existing=True,tempdir=None,osmconvert_path='osmconvert',mapping=None):
        self.hostname = hostname
        self._path = path
        self.geom = geom
        self.use_existing = use_existing
        self.osmconvert_path = osmconvert_path
        self.tmp_path = os.path.join(tempdir,'tmp.osm.xml')
        self.mapping = mapping

    def fetch(self):
        base_template = Template('[maxsize:$maxsize][timeout:$timeout];$query;out meta;')

        if self.geom.geom_type == 'Polygon':
            geom = 'poly:"{0}"'.format(' '.join(['{1} {0}'.format(*x) for x in self.geom.exterior.coords]))
        else:
            bounds = self.geom.bounds
            west = max(bounds[0], -180)
            south = max(bounds[1], -90)
            east = min(bounds[2], 180)
            north = min(bounds[3], 90)
            geom = '{1},{0},{3},{2}'.format(west, south, east, north)

        if self.mapping:
            query = """(
                (
                    {0}
                );
                (
                    {1}
                );>;
                (
                    {2}
                );>>;>;)"""
            nodes,ways,relations = Overpass.filters(self.mapping)
            nodes = '\n'.join(['node({0}){1};'.format(geom,f) for f in nodes])
            ways = '\n'.join(['way({0}){1};'.format(geom,f) for f in ways])
            relations = '\n'.join(['relation({0}){1};'.format(geom,f) for f in relations])
            query = query.format(nodes,ways,relations)
        else:
            query = '(node({0});<;>>;>;)'.format(geom)

        data = base_template.substitute(maxsize=2147483648,timeout=1600,query=query)

        with requests.post(os.path.join(self.hostname,'api','interpreter'),data=data, stream=True) as r:
            with open(self.tmp_path, 'wb') as f:
                shutil.copyfileobj(r.raw, f)

        with open(self.tmp_path,'r') as f:
            sample = [next(f) for x in range(6)]
            if 'DOCTYPE html' in sample[1]:
                raise Exception('Overpass failure')
            if 'remark' in sample[5]:
                raise Exception(sample[5])

        # run osmconvert on the file
        subprocess.check_call([self.osmconvert_path,self.tmp_path,'--out-pbf','-o='+self._path])
        os.remove(self.tmp_path)

    def path(self):
        if os.path.isfile(self._path) and self.use_existing:
            return self._path
        else:
            self.fetch()
        return self._path





import json
import os
from os.path import basename
import zipfile
import tarfile
import io
from shapely.geometry import mapping
from osm_export_tool import File

def create_package(destination,files,boundary_geom=None,output_name='zip'):
    with zipfile.ZipFile(destination, 'w', zipfile.ZIP_DEFLATED, True) as z:
        if boundary_geom:
            z.writestr("clipping_boundary.geojson", json.dumps(mapping(boundary_geom)))
        for file in files:
            for part in file.parts:
                z.write(part, os.path.basename(part))

    return File(output_name,[destination])

def create_posm_bundle(destination,files,title,name,description,geom):
    contents = {}
    with tarfile.open(destination, "w|gz") as bundle:
        for file in files:
            for part in file.parts:
                if file.output_name == 'shp':
                    target = 'data/' + basename(part)
                    contents[target] = {'Type':'ESRI Shapefile'}
                elif file.output_name == 'kml':
                    target = 'data/' + basename(part)
                    contents[target] = {'Type':'KML'}
                elif file.output_name == 'gpkg':
                    target = 'data/' + basename(part)
                    contents[target] = {'Type':'Geopackage'}
                elif file.output_name == 'osmand_obf':
                    target = 'navigation/' + basename(part)
                    contents[target] = {'Type':'OsmAnd'}
                elif file.output_name == 'garmin':
                    target = 'navigation/' + basename(part)
                    contents[target] = {'Type':'Garmin IMG'}
                elif file.output_name == 'mwm':
                    target = 'navigation/' + basename(part)
                    contents[target] = {'Type':'Maps.me'}
                elif file.output_name == 'osm_pbf':
                    target = 'osm/' + basename(part)
                    contents[target] = {'Type':'OSM/PBF'}
                elif file.output_name == 'mbtiles':
                    target = 'tiles/' + basename(part)
                    contents[target] = {
                        'type':'MBTiles',
                        'minzoom':file.extra['minzoom'],
                        'maxzoom':file.extra['maxzoom'],
                        'source':file.extra['source']
                    }
                bundle.add(part,target)

        data = json.dumps({
            'title':title,
            'name':name,
            'description':description,
            'bbox':geom.bounds,
            'contents':contents
        },indent=2).encode()
        tarinfo = tarfile.TarInfo('manifest.json')
        tarinfo.size = len(data)
        bundle.addfile(tarinfo, io.BytesIO(data))

    return File('bundle',[destination])



import yaml
import pyparsing
from osm_export_tool import GeomType
from osm_export_tool.sql import Matcher

class InvalidMapping(Exception):
	pass

class Theme:
	def __init__(self,name,d,default_osm_id):
		self.name = name

		# set geometry types.
		self.points = False
		self.lines = False
		self.polygons = False

		if 'types' not in d:
			self.points = True
			self.lines = True
			self.polygons = True
		else:
			for t in d['types']:
				if t not in ['points','lines','polygons']:
					raise InvalidMapping('types: for theme {0} must be list containing one or more of: points, lines, polygons'.format(name))
			if 'points' in d['types']:
				self.points = True
			if 'lines' in d['types']:
				self.lines = True
			if 'polygons' in d['types']:
				self.polygons = True

		if isinstance(d,list):
			raise InvalidMapping('theme {0} must be YAML dict (types: , select:) , not list (- types, - select)'.format(name))

		if 'select' not in d:
			raise InvalidMapping('missing select: for theme {0}'.format(name))
		self.keys = set(d['select'])

		self.osm_id = default_osm_id
		if 'osm_id' in self.keys:
			self.osm_id = True
			self.keys.remove('osm_id')

		if 'where' in d:
			if not d['where']:
				raise InvalidMapping('where: for theme {0} is invalid'.format(name))
			if isinstance(d['where'],list):
				self.matcher = Matcher.null()
				for w in d['where']:
					self.matcher = self.matcher.union(Matcher.from_sql(w))
			else:
				try:
					self.matcher = Matcher.from_sql(d['where'])
				except pyparsing.ParseException:
					raise InvalidMapping('Invalid SQL: {0}'.format(d['where']))
		else:
			self.matcher = Matcher.null()
			for key in self.keys:
				self.matcher = self.matcher.union(Matcher.any(key))

		extra = d.copy()
		if 'where' in extra:
			del extra['where']
		if 'select' in d:
			del extra['select']
		if 'types' in d:
			del extra['types']
		self.extra = extra

	def matches(self,geom_type,tags):
		if geom_type == GeomType.POINT and not self.points:
			return False
		if geom_type == GeomType.LINE and not self.lines:
			return False
		if geom_type == GeomType.POLYGON and not self.polygons:
			return False

		return self.matcher.matches(tags)

	def __repr__(self):
		return self.name


class Mapping:
	def __init__(self,y,default_osm_id=True):
		doc = yaml.safe_load(y)

		if not isinstance(doc,dict):
			raise InvalidMapping('YAML must be dict')

		self.themes = []
		for theme_name, theme_dict in doc.items():
			self.themes.append(Theme(theme_name,theme_dict,default_osm_id=default_osm_id))

	@classmethod
	def validate(cls,y,**kwargs):
		try:
			return cls(y,kwargs), None
		except (yaml.scanner.ScannerError, yaml.parser.ParserError, InvalidMapping) as se:
			errors = [str(se)]
			return None, errors




import os
from os.path import join
import pathlib
import subprocess
from osm_export_tool import File
import landez

def osmand(input_pbf,map_creator_dir,tempdir=None,jvm_mem=[256,2048]):
    BATCH_XML = """<?xml version="1.0" encoding="utf-8"?>
        <batch_process>
            <process_attributes mapZooms="" renderingTypesFile="" zoomWaySmoothness=""
                osmDbDialect="sqlite" mapDbDialect="sqlite"/>
             <!-- zoomWaySmoothness - 1-4, typical mapZooms - 11;12;13-14;15-   -->
            <process directory_for_osm_files="{tempdir}/osmand"
                     directory_for_index_files="{tempdir}"
                     directory_for_generation="{tempdir}"
                     skipExistingIndexesAt="{tempdir}"
                     indexPOI="true"
                     indexRouting="true"
                     indexMap="true"
                     indexTransport="true"
                     indexAddress="true">
            </process>
        </batch_process>
        """
    CLASSPATH = "{map_creator_dir}/OsmAndMapCreator.jar:{map_creator_dir}/lib/OsmAnd-core.jar:{map_creator_dir}/lib/*.jar"

    pathlib.Path(join(tempdir,'osmand')).mkdir(parents=True, exist_ok=True)

    try:
        os.link(input_pbf,join(tempdir,'osmand','osmand.osm.pbf'))
    except:
        pass

    with open(join(tempdir,'batch.xml'),'w') as b:
        b.write(BATCH_XML.format(tempdir=tempdir))

    subprocess.check_call([
        'java',
        '-Xms{0}M'.format(jvm_mem[0]),
        '-Xmx{0}M'.format(jvm_mem[1]),
        '-cp',
        CLASSPATH.format(map_creator_dir=map_creator_dir),
        'net.osmand.util.IndexBatchCreator',
        join(tempdir,'batch.xml')
    ])
    return [File('osmand_obf',[join(tempdir,'Osmand_2.obf')])]

def garmin(input_pbf,splitter_jar,mkgmap_jar,tempdir=None,jvm_mem=[256,2048]):
    """
    Converts PBF to Garmin IMG format.

    Splits pbf into smaller tiles, generates .img files for each split,
    then patches the .img files back into a single .img file
    suitable for deployment to a Garmin GPS unit.
    NOTE: disabled poly bounds: see https://github.com/hotosm/osm-export-tool2/issues/248
    """
    subprocess.check_call([
        'java',
        '-Xms{0}M'.format(jvm_mem[0]),
        '-Xmx{0}M'.format(jvm_mem[1]),
        '-jar',
        splitter_jar,
        '--output-dir=' + tempdir,
        input_pbf
    ])

    # Generate the IMG file.
    # get the template.args file created by splitter
    # see: http://wiki.openstreetmap.org/wiki/Mkgmap/help/splitter
    subprocess.check_call([
        'java',
        '-Xms{0}M'.format(jvm_mem[0]),
        '-Xmx{0}M'.format(jvm_mem[1]),
        '-jar',
        mkgmap_jar,
        '--gmapsupp',
        '--output-dir=' + tempdir,
        '--description="HOT Export Garmin Map"',
        '--mapname=80000111',
        '--family-name="HOT Export Tool"',
        '--family-id="2"',
        '--series-name="HOT Export Tool"',
        '--index',
        '--route',
        '--generate-sea=extend-sea-sectors',
        '--draw-priority=100',
        '--read-config={0}/template.args'.format(tempdir)
    ])
    return [File('garmin',[join(tempdir,'gmapsupp.img')])]

def mwm(input_pbf,output_dir,generate_mwm_path,generator_tool_path,osmconvert_path='osmconvert'):
    base_name = (os.path.basename(input_pbf).split(os.extsep))[0]
    env = os.environ.copy()
    env.update(OSMCONVERT=osmconvert_path,TARGET=output_dir,GENERATOR_TOOL=generator_tool_path)
    subprocess.check_call([
        generate_mwm_path,
        input_pbf
    ],env=env)
    return [File('mwm',[join(output_dir,base_name + '.mwm')])]

def mbtiles(geom,filepath,tiles_url,minzoom,maxzoom):
    mb = landez.MBTilesBuilder(cache=False,tiles_url=tiles_url,tiles_headers={'User-Agent':'github.com/hotosm/osm-export-tool'},filepath=filepath)
    mb.add_coverage(bbox=geom.bounds,
                    zoomlevels=[minzoom,maxzoom])
    mb.run()
    return [File('mbtiles',[filepath],{'minzoom':minzoom,'maxzoom':maxzoom,'source':tiles_url})]



import unittest
from osm_export_tool.sources import Overpass
from osm_export_tool.mapping import Mapping

class TestMappingToOverpass(unittest.TestCase):
    def test_mapping(self):
        y = '''
        buildings:
            types:
                - points
            select:
                - column1 
            where: column2 IS NOT NULL

        other1:
            types:
                - points
                - polygons
            select:
                - column1 
                - irrelevant
            where: column2 IS NOT NULL AND column3 IN ('foo','bar')

        other2:
            types:
                - lines
            select:
                - column5:key
        '''
        mapping = Mapping(y)
        nodes, ways, relations = Overpass.filters(mapping)
        self.assertCountEqual(nodes,["['column3'~'foo|bar']","['column2']"])
        # force quoting of strings to handle keys with colons
        self.assertCountEqual(ways,["['column5:key']","['column3'~'foo|bar']","['column2']"])
        self.assertCountEqual(relations,["['column3'~'foo|bar']","['column2']"])

class TestSQLToOverpass(unittest.TestCase):
    def test_basic(self):
        s = Overpass.sql("name = 'somename'")
        self.assertEqual(s,["['name'='somename']"])
        s = Overpass.sql("level > 4")
        self.assertEqual(s,["['level']"])

    def test_basic_list(self):
        s = Overpass.sql("name IN ('val1','val2')")
        self.assertEqual(s,["['name'~'val1|val2']"])

    def test_whitespace(self):
        s = Overpass.sql("name = 'some value'")
        self.assertEqual(s,["['name'='some value']"])

    def test_notnull(self):
        s = Overpass.sql("name is not null")
        self.assertEqual(s,["['name']"])

    def test_and_or(self):
        s = Overpass.sql("name1 = 'foo' or name2 = 'bar'")
        self.assertEqual(s,["['name1'='foo']","['name2'='bar']"])
        s = Overpass.sql("(name1 = 'foo' and name2 = 'bar') or name3 = 'baz'")
        self.assertEqual(s,["['name1'='foo']","['name2'='bar']","['name3'='baz']"])


import setuptools
import subprocess
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = [
    'osmium~=2.15',
    'pyparsing~=2.4',
    'pyyaml',
    'shapely~=1.6',
    'requests~=2.22.0',
    'landez~=2.5.0'
]

if 'GDAL_VERSION' in os.environ:
    gdal_version = os.environ['GDAL_VERSION']
else:
    gdal_config = os.environ.get('GDAL_CONFIG', 'gdal-config')
    gdal_version = subprocess.check_output([gdal_config,'--version'],encoding='UTF-8').strip()

requirements.append(['gdal~=' + gdal_version])

setuptools.setup(
    name="osm-export-tool",
    version="0.0.14",
    author="Brandon Liu",
    author_email="brandon.liu@hotosm.org",
    description="Convert OpenStreetMap data into GIS and mobile mapping file formats.",
    license="BSD-3-Clause",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hotosm/osm-export-tool-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    scripts=['bin/osm-export-tool'],
    install_requires = requirements,
    requires_python='>=3.0',
    package_data={'osm_export_tool':['mappings/*.yml']}
)


from pyparsing import Word, delimitedList, Optional, \
    Group, alphas, nums, alphanums, ParseException, Forward, oneOf, quotedString, \
    ZeroOrMore, Keyword 

class InvalidSQL(Exception):
    pass


ident          = Word( alphas, alphanums + "_:" )
columnName =   (ident | quotedString())("columnName")

whereExpression = Forward()
and_ = Keyword("and", caseless=True)('and')
or_ = Keyword("or", caseless=True)('or')
in_ = Keyword("in", caseless=True)("in")
isnotnull = Keyword("is not null",caseless=True)('notnull')
binop = oneOf("= != < > >= <=", caseless=True)('binop')
intNum = Word( nums )

columnRval = (intNum | quotedString.setParseAction(lambda x:x[0][1:-1]))('rval*')
whereCondition = Group(
    ( columnName + isnotnull ) |
    ( columnName + binop + columnRval ) |
    ( columnName + in_ + "(" + delimitedList( columnRval ) + ")" ) |
    ( "(" + whereExpression + ")" )
    )('condition')
whereExpression << Group(whereCondition + ZeroOrMore( ( and_ | or_ ) + whereExpression ) )('expression')

class SQLValidator(object):
    """ Parses a subset of SQL to define feature selections.
        This validates the SQL to make sure the user can't do anything dangerous."""

    def __init__(self,s):
        self._s = s
        self._errors = []
        self._parse_result = None

    @property
    def valid(self):
        try:
            self._parse_result = whereExpression.parseString(self._s,parseAll=True)
        except InvalidSQL as e:
            self._errors.append(str(e))
            return False
        except ParseException as e:
            self._errors.append("SQL could not be parsed.")
            return False
        return True

    @property
    def errors(self):
        return self._errors

    @property
    def column_names(self):
        # takes a dictionary, returns a list
        def column_names_in_dict(d):
            result = []
            for key, value in d.items():
                if 'columnName'  == key:
                    result = result + [value]
                if isinstance(value,dict):
                    result = result + column_names_in_dict(value)
            return result
        return column_names_in_dict(self._parse_result.asDict())

def strip_quotes(token):
    if token[0] == '"' and token[-1] == '"':
        token = token[1:-1]
    if token[0] == "'" and token[-1] == "'":
        token = token[1:-1]
    return token

def _match(d,tags):
    if len(d) == 0:
        return False
    op = d[0]
    if op == 'or':
        return _match(d[1],tags) or _match(d[2],tags)
    elif op == 'and':
        return _match(d[1],tags) and _match(d[2],tags)
    elif op == '=':
        return d[1] in tags and tags[d[1]] == d[2]
    elif op == 'notnull':
        return d[1] in tags
    elif op == 'in':
        return (d[1] in tags) and (tags[d[1]] in d[2])
    elif op == '!=':
        return d[1] not in tags or tags[d[1]] != d[2]
    elif op == '>':
        return d[1] in tags and str(tags[d[1]]) > str(d[2])
    elif op == '<':
        return d[1] in tags and str(tags[d[1]]) < str(d[2])
    elif op == '>=':
        return d[1] in tags and str(tags[d[1]]) >= str(d[2])
    elif op == '<=':
        return d[1] in tags and str(tags[d[1]]) <= str(d[2])
    raise Exception

def to_prefix(sql):
    def prefixform(d):
        if 'or' in d:
            return ('or',prefixform(d['condition']),prefixform(d['expression']))
        elif 'and' in d:
            return ('and',prefixform(d['condition']),prefixform(d['expression']))
        elif 'condition' in d:
            return prefixform(d['condition'])
        elif 'expression' in d:
            return prefixform(d['expression'])
        elif 'binop' in d:
            return (d['binop'],strip_quotes(d['columnName']),d['rval'][0])
        elif 'notnull' in d:
            return ('notnull',strip_quotes(d['columnName']))
        elif 'in' in d:
            return ('in',strip_quotes(d['columnName']),d['rval'])
    return prefixform(whereExpression.parseString(sql,parseAll=True).asDict())

class Matcher:
    def __init__(self,expr):
        self.expr = expr

    def matches(self,tags):
        return _match(self.expr,tags)

    # returns a new matcher
    def union(self,other_matcher):
        if other_matcher.expr == ():
            return Matcher(self.expr)
        if self.expr == ():
            return Matcher(other_matcher.expr)
        return Matcher(('or',self.expr,other_matcher.expr))

    @classmethod
    def any(cls,tag_name):
        return Matcher(('notnull',tag_name))

    @classmethod
    def null(cls):
        return Matcher(())

    @classmethod
    def from_sql(cls,sql):
        return cls(to_prefix(sql))

    # only used for display and debugging
    def to_sql(self):
        def expr_to_sql(e):
            if e[0] == '=':
                return "{0} = '{1}'".format(e[1],e[2])
            if e[0] == 'notnull':
                return "{0} IS NOT NULL".format(e[1])
            if e[0] == '!=':
                return "{0} != '{1}'".format(e[1],e[2])
            if e[0] == '>=':
                return "{0} >= {1}".format(e[1],e[2])
            if e[0] == '<=':
                return "{0} <= {1}".format(e[1],e[2])
            if e[0] == '>':
                return "{0} > {1}".format(e[1],e[2])
            if e[0] == '<':
                return "{0} < {1}".format(e[1],e[2])
            if e[0] == 'in':
                parts = ','.join(["'" + x + "'" for x in e[2]])
                return '{0} IN ({1})'.format(e[1],parts)
            if e[0] == 'and':
                return expr_to_sql(e[1]) + ' AND ' + expr_to_sql(e[2])
            if e[0] == 'or':
                return expr_to_sql(e[1]) + ' OR ' + expr_to_sql(e[2])
        return expr_to_sql(self.expr)



import unittest
from osm_export_tool.sql import SQLValidator, Matcher

class TestSql(unittest.TestCase):

    def test_basic(self):
        s = SQLValidator("name = 'a name'")
        self.assertTrue(s.valid)

    def test_identifier_list(self):
        s = SQLValidator("natural in ('water','cliff')")
        self.assertTrue(s.valid)

    def test_float_value(self):
        s = SQLValidator("height > 20")
        self.assertTrue(s.valid)

    def test_not_null(self):
        s = SQLValidator("height IS NOT NULL")
        self.assertTrue(s.valid)

    def test_and_or(self):
        s = SQLValidator("height IS NOT NULL and height > 20")
        self.assertTrue(s.valid)
        s = SQLValidator("height IS NOT NULL or height > 20")
        self.assertTrue(s.valid)
        s = SQLValidator("height IS NOT NULL or height > 20 and height < 30")
        self.assertTrue(s.valid)

    def test_parens(self):
        s = SQLValidator("(admin IS NOT NULL and level > 4)")
        self.assertTrue(s.valid)
        s = SQLValidator("(admin IS NOT NULL and level > 4) AND height is not null")
        self.assertTrue(s.valid)

    def test_colons_etc(self):
        s = SQLValidator("addr:housenumber IS NOT NULL")
        self.assertTrue(s.valid)
        s = SQLValidator("admin_level IS NOT NULL")
        self.assertTrue(s.valid)

    def test_invalid_sql(self):
        s = SQLValidator("drop table planet_osm_polygon")
        self.assertFalse(s.valid)
        self.assertEqual(s.errors,['SQL could not be parsed.'])
        s = SQLValidator("(drop table planet_osm_polygon)")
        self.assertFalse(s.valid)
        self.assertEqual(s.errors,['SQL could not be parsed.'])
        s = SQLValidator ("")
        self.assertFalse(s.valid)
        self.assertEqual(s.errors,['SQL could not be parsed.'])
        s = SQLValidator("name = 'a name'; blah")
        self.assertFalse(s.valid)
        self.assertEqual(s.errors,['SQL could not be parsed.'])

    def test_column_names(self):
        s = SQLValidator("(admin IS NOT NULL and level > 4) AND height is not null")
        self.assertTrue(s.valid)
        self.assertEqual(s.column_names,['admin','level','height'])

class TestMatcher(unittest.TestCase):
    def test_matcher_binop(self):
        m = Matcher.from_sql("building = 'yes'")
        self.assertTrue(m.matches({'building':'yes'}))
        self.assertFalse(m.matches({'building':'no'}))

        m = Matcher.from_sql("building != 'yes'")
        self.assertFalse(m.matches({'building':'yes'}))
        self.assertTrue(m.matches({'building':'no'}))

    def test_matcher_colon(self):
        m = Matcher.from_sql("addr:housenumber = 1")
        self.assertTrue(m.matches({'addr:housenumber':'1'}))

        m = Matcher.from_sql("building != 'yes'")
        self.assertFalse(m.matches({'building':'yes'}))
        self.assertTrue(m.matches({'building':'no'}))

    def test_matcher_doublequote(self):
        m = Matcher.from_sql("\"addr:housenumber\" = 1")
        self.assertTrue(m.matches({'addr:housenumber':'1'}))

        m = Matcher.from_sql("\"addr:housenumber\" IN ('foo')")
        self.assertTrue(m.matches({'addr:housenumber':'foo'}))

        m = Matcher.from_sql("\"addr:housenumber\" IS NOT NULL")
        self.assertTrue(m.matches({'addr:housenumber':'foo'}))

    def test_matcher_or(self):
        m = Matcher.from_sql("building = 'yes' OR amenity = 'bank'")
        self.assertTrue(m.matches({'building':'yes'}))
        self.assertTrue(m.matches({'amenity':'bank'}))
        self.assertFalse(m.matches({}))

    def test_matcher_and(self):
        m = Matcher.from_sql("building = 'yes' AND amenity = 'bank'")
        self.assertFalse(m.matches({'building':'yes'}))
        self.assertFalse(m.matches({'amenity':'bank'}))

    def test_matcher_is_not_null(self):
        m = Matcher.from_sql("building IS NOT NULL")
        self.assertTrue(m.matches({'building':'one'}))
        self.assertTrue(m.matches({'building':'two'}))
        self.assertFalse(m.matches({}))

    def test_in(self):
        m = Matcher.from_sql("building IN ('one','two')")
        self.assertTrue(m.matches({'building':'one'}))
        self.assertTrue(m.matches({'building':'two'}))
        self.assertFalse(m.matches({}))
        self.assertFalse(m.matches({'building':'three'}))

    def test_any(self):
        m = Matcher.any("building");
        self.assertTrue(m.matches({'building':'one'}))

    def test_union(self):
        m = Matcher.any("building").union(Matcher.any("parking"))
        self.assertTrue(m.matches({'building':'one'}))
        self.assertTrue(m.matches({'parking':'one'}))

    def test_null(self):
        m = Matcher.null()
        self.assertFalse(m.matches({'building':'one'}))

    def test_to_sql(self):
        sql = "building = 'yes'"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building IS NOT NULL"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building IN ('one','two')"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building != 'yes'"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building >= 0"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building <= 0"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building > 0"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building < 0"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building > 0 AND building < 5"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
        sql = "building > 0 OR building < 5"
        self.assertEqual(Matcher.from_sql(sql).to_sql(),sql)
